;; Generated package description from hiddenquote.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "hiddenquote" "1.2.0.20250305.95113" "Major mode for doing hidden quote puzzles" '((emacs "25.1")) :commit "743e815a07b40eea5b1f65ef5f753922cadaaea0" :authors '(("Mauro Aranda" . "maurooaranda@gmail.com")) :maintainer '("Mauro Aranda" . "maurooaranda@gmail.com") :keywords '("games") :url "https://mauroaranda.com/puzzles/hidden-quote-puzzle/")
